import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
from matplotlib import patheffects
import trieste
from trieste.space import Box
from trieste.data import Dataset
from trieste.models.gpflow import build_gpr, GaussianProcessRegression
from trieste.acquisition.combination import Product
from trieste.acquisition.rule import EfficientGlobalOptimization

def BBO_model(seedVal, initPoints, numSteps):

    np.random.seed(seedVal)
    tf.random.set_seed(seedVal)


    def f(x):
        if x.size == 2:
            x1 = x[0]
            x2 = x[1]
        else:
            x1 = x[:,0]
            x2 = x[:,1]
        
        a = 15*x2 - 5.1*((15*x1-5)**2)/(4*np.pi**2) + (75*x1-25)/(np.pi) - 6

        b = 1 - (np.cos(15*x1-4))/(8*np.pi) + 75*x1 - 25

        return a**2 + 10*b

    def g(x):
        if x.size == 2:
            x1 = x[0]
            x2 = x[1]
        else:
            x1 = x[:,0]
            x2 = x[:,1]

        return (10-2*x1**2+x1**4/3)*x1**2 + x1*x2 + (4*x2**2-4)*x2**2 + 4*np.sin(5*np.pi*(1-x1)) + 4*np.sin(6*np.pi*(1-x2)) - 6

    def h(x):
        if x.size == 2:
            x1 = x[0]
            x2 = x[1]
        else:
            x1 = x[:,0]
            x2 = x[:,1]

        return 20*(x1-0.7)**2 - 0.25 - x2

    class Sim:
        threshold = 0
        threshold2 = 1e-4
        threshold3 = 1e-4

        @staticmethod
        def objective(input_data):
            x1, x2 = input_data[...,-2], input_data[...,-1]

            a = 15*x2 - 5.1*((15*x1-5)**2)/(4*np.pi**2) + (75*x1-25)/(np.pi) - 6
            b = 1 - (tf.cos(15*x1-4))/(8*np.pi) + 75*x1 - 25

            z = a**2 + 10*b
            return z[:,None]

        @staticmethod
        def constraint(input_data):
            x1, x2 = input_data[...,-2], input_data[...,-1]

            z = (10-2*x1**2+x1**4/3)*x1**2 + x1*x2 + (4*x2**2-4)*x2**2 + 4*tf.sin(5*np.pi*(1-x1)) + 4*tf.sin(6*np.pi*(1-x2)) - 6
            return z[:,None]
        
        @staticmethod
        def constraint2(input_data):
            x1, x2 = input_data[...,-2], input_data[...,-1]

            z = 20*(x1-0.7)**2 - 0.25 - x2
            return z[:,None]

        @staticmethod
        def constraint3(input_data):
            x1, x2 = input_data[...,-2], input_data[...,-1]

            z = 20*(x1-0.7)**2 - 0.25 - x2
            return -z[:,None]

    search_space = Box([0,0],[1,1])


    OBJECTIVE = "OBJECTIVE"
    CONSTRAINT = "CONSTRAINT"
    CONSTRAINT2 = "CONSTRAINT2"
    CONSTRAINT3 = "CONSTRAINT3"

    def observer(query_points):
        return {
            OBJECTIVE: Dataset(query_points, Sim.objective(query_points)),
            CONSTRAINT: Dataset(query_points, Sim.constraint(query_points)),
            CONSTRAINT2: Dataset(query_points, Sim.constraint2(query_points)),
            CONSTRAINT3: Dataset(query_points, Sim.constraint3(query_points)),
        }

    def create_bo_model(data):
        gpr = build_gpr(data, search_space, likelihood_variance=1e-7)
        return GaussianProcessRegression(gpr)

    num_initial_points = initPoints
    initial_data = observer(search_space.sample(num_initial_points))
    initial_models = trieste.utils.map_values(create_bo_model, initial_data)

    pof1 = trieste.acquisition.ProbabilityOfFeasibility(threshold=Sim.threshold)
    pof2 = trieste.acquisition.ProbabilityOfFeasibility(threshold=Sim.threshold2)
    pof3 = trieste.acquisition.ProbabilityOfFeasibility(threshold=Sim.threshold3)
    pof = Product(pof1.using(CONSTRAINT), pof2.using(CONSTRAINT2), pof3.using(CONSTRAINT3))

    eci = trieste.acquisition.ExpectedConstrainedImprovement(OBJECTIVE, pof)  # type: ignore
    rule = EfficientGlobalOptimization(eci)

    num_steps = numSteps
    bo = trieste.bayesian_optimizer.BayesianOptimizer(
        observer, search_space
    )

    data = bo.optimize(
        num_steps, initial_data, initial_models, rule, track_state=True
    ).try_get_final_datasets()



    mask_fail1 = (
        data[CONSTRAINT].observations.numpy().flatten().astype(float) < Sim.threshold
    )
    mask_fail2 = (
        data[CONSTRAINT2].observations.numpy().flatten().astype(float)
        > Sim.threshold2
    )
    mask_fail3 = (
        data[CONSTRAINT3].observations.numpy().flatten().astype(float)
        > Sim.threshold3
    )

    mask_fail = np.logical_or(mask_fail1, mask_fail2, mask_fail3)

    resultX = []
    resultY = []
    gvals   = []
    hvals   = []

    for i in range(mask_fail.size):
        Xval = data[OBJECTIVE].query_points.numpy()[i]
        Yval = data[OBJECTIVE].observations.numpy()[i]
        if not mask_fail[i]:
            resultX.append(Xval)
            resultY.append(Yval)
            gvals.append(g(Xval))
            hvals.append(h(Xval))

    resultX = [x for _, x in sorted(zip(resultY, resultX))]
    gvals   = [x for _, x in sorted(zip(resultY, gvals))]
    hvals   = [x for _, x in sorted(zip(resultY, hvals))]
    resultY = sorted(resultY)

    return resultX, gvals, hvals, resultY

resultX, gvals, hvals, resultY = BBO_model(1404,10,15)

print('From B-Opt:')
print(resultY[0])
print(resultX[0])
print('g: ', gvals[0])
print('h: ', hvals[0])