from casadi import *
nlpopts = {'ipopt': {'tol':1e-12, 'print_level': 0}, 'print_time': False}

class DynModel:
    def __init__(self):
        self.m = {
            't': MX.sym('t'),
            'x': DM([]),
            'z': DM([]),
            'u': DM([]),
            'p': DM([]),
            'f': DM([]),
            'h': DM([]),
            'g': DM([]),
            'J': DM([])
        }
        self.jacs = None
        # self.jacs = {
        #     'dfdu': DM([]),
        #     'dfdx': DM([]),
        #     'dJdu': DM([]),
        #     'dJdx': DM([]),
        #     'dgdu': DM([]),
        #     'dgdx': DM([])
        # }

    def guess(self, d):
        nx = self.m['x'].shape
        nz = self.m['z'].shape
        nu = self.m['u'].shape

        x0  = DM.zeros(nx)
        u0  = DM.zeros(nu)
        z0  = DM.zeros(nz)
        lbx = -DM.inf(nx)
        lbu = -DM.inf(nu)
        lbz = -DM.inf(nz)
        ubx = DM.inf(nx)
        ubu = DM.inf(nu)
        ubz = DM.inf(nz)
        return u0, lbu, ubu, x0, lbx, ubx, z0, lbz, ubz

    def gen_jacs(self):
        xz = vertcat(self.m['x'], self.m['z'])
        fh = vertcat(self.m['f'], self.m['h'])
        self.jacs = {
            'dfdu': jacobian(fh, self.m['u']),
            'dfdp': jacobian(fh, self.m['p']),
            'dfdx': jacobian(fh, xz),
        }
        if 'g' in self.m:
            self.jacs |= {
                'dJdu': jacobian(self.m['J'], self.m['u']),
                'dJdp': jacobian(self.m['J'], self.m['p']),
                'dJdx': jacobian(self.m['J'], xz),
            }
        if 'g' in self.m:
            self.jacs |= {
                'dgdu': jacobian(self.m['g'], self.m['u']),
                'dgdp': jacobian(self.m['g'], self.m['p']),
                'dgdx': jacobian(self.m['g'], xz),
            }
        if 'y' in self.m:
            self.jacs |= {
                'dydu': jacobian(self.m['y'], self.m['u']),
                'dydp': jacobian(self.m['y'], self.m['p']),
                'dydx': jacobian(self.m['y'], xz),
            }

    def linearize(self, point: dict):
        lin_mod = DynModel()
        if self.jacs is None: self.gen_jacs()
        xz = vertcat(self.m['x'], self.m['z'])
        nx = self.m['x'].shape[0]

        df_fun = Function('df_fun', [xz, self.m['u'], self.m['p']], [self.jacs['dfdu'], self.jacs['dfdp'], self.jacs['dfdx']])
        B, Bd, A = df_fun(point['xz'], point['u'], point['p'])
        fh = vertcat(self.m['f'], self.m['h'])
        fh0_fun = Function('fh0_fun', [xz, self.m['u'], self.m['p']], [fh])
        fh0 = fh0_fun(point['xz'], point['u'], point['p'])
        fhlin = fh0 + A @ (xz - point['xz']) + B @ (self.m['u'] - point['u']) + Bd @ (self.m['p'] - point['p'])
        lin_mod.m |= {
            't': self.m['t'],
            'x': self.m['x'],
            'z': self.m['z'],
            'u': self.m['u'],
            'p': self.m['p'],
            'f': fhlin[:nx],
            'h': fhlin[nx:],
        }
        if 'g' in self.m:
            dg_fun = Function('dg_fun', [xz, self.m['u'], self.m['p']], [self.jacs['dgdu'], self.jacs['dgdp'], self.jacs['dgdx']])
            gB, gBd, gA = dg_fun(point['xz'], point['u'], point['p'])
            g0_fun = Function('g0_fun', [xz, self.m['u'], self.m['p']], [self.m['g']])
            g0 = g0_fun(point['xz'], point['u'], point['p'])
            glin = g0 + gA @ (xz - point['xz']) + gB @ (self.m['u'] - point['u']) + gBd @ (self.m['p'] - point['p'])
            lin_mod.m |= {'g': glin}
        # if 'J' in self.m:
        #     dJ_fun = Function('dJ_fun', [xz, self.m['u'], self.m['p']], [self.jacs['dJdu'], self.jacs['dJdp'], self.jacs['dJdx']])
        #     JB, JBd, JA = dJ_fun(point['xz'], point['u'], point['p'])
        #     J0_fun = Function('J0_fun', [xz, self.m['u'], self.m['p']], [self.m['J']])
        #     J0 = J0_fun(point['xz'], point['u'], point['p'])
        #     Jlin = J0 + JA @ (xz - point['xz']) + JB @ (self.m['u'] - point['u']) + JBd @ (self.m['p'] - point['p'])
        #     lin_mod.m |= {'J': Jlin}
        if 'y' in self.m:
            dy_fun = Function('dy_fun', [xz, self.m['u'], self.m['p']], [self.jacs['dydu'], self.jacs['dydp'], self.jacs['dydx']])
            yB, yBd, yA = dy_fun(point['xz'], point['u'], point['p'])
            y0_fun = Function('y0_fun', [xz, self.m['u'], self.m['p']], [self.m['y']])
            y0 = y0_fun(point['xz'], point['u'], point['p'])
            ylin = y0 + yA @ (xz - point['xz']) + yB @ (self.m['u'] - point['u']) + yBd @ (self.m['p'] - point['p'])
            lin_mod.m |= {'y': ylin}
        
        return lin_mod
    
    def simulate(self, x0, t_span, u, d, z0=[]):
        dae = {
            'x': self.m['x'],
            'p': vertcat(self.m['u'], self.m['p']),
            'ode': self.m['f']
        }
        if 't' in self.m: dae['t'] = self.m['t']
        if 'z' in self.m: dae['z'] = self.m['z']
        if 'h' in self.m: dae['alg'] = self.m['h']
        if 'J' in self.m: dae['quad'] = self.m['J']

        opts = {
            't0': t_span[0],
            'tf': t_span[-1],
            'grid': t_span,
            'output_t0': True
        }
        F = integrator('F', 'idas', dae, opts)
        r = F(x0=x0, z0=z0, p=vertcat(u,d))
        r['tf'] = DM(t_span)

        if 'y' in self.m:
            yfun = Function('yf', [self.m['t'], self.m['x'], self.m['u'], self.m['p']], [self.m['y']])
            y = [ yfun(ti, r['xf'][:,i], u, d) for i,ti in enumerate(t_span) ]
            r['yf'] = hcat(y)

        return r

    def ss_grad(self, u, d, x=None):
        m = self.m
        if hasattr(self, 'jacs'):
            if self.jacs is None:
                self.gen_jacs()
        else:
            self.gen_jacs()
        jacs = self.jacs
        xz = vertcat(self.m['x'], self.m['z'])
        fh = vertcat(self.m['f'], self.m['h'])
        jacs_fun = Function('Ju', [xz, m['u'], m['p']], [jacs['dfdu'], jacs['dfdx'], jacs['dgdu'], jacs['dgdx'], jacs['dJdu'], jacs['dJdx']])
        # Jufun = Function('Ju', [xz, m['u'], m['p']], [jacs['dJdu'] - jacs['dJdx'] @ solve(jacs['dfdx'], jacs['dfdu'])])
        # gufun = Function('gu', [xz, m['u'], m['p']], [jacs['dgdu'] - jacs['dgdx'] @ solve(jacs['dfdx'], jacs['dfdu'])])
        flag = True
        if x is None:
            out = self.ss_output(u,d) #np.zeros(nx)
            x = out['xz']
            flag = out['success']
        dfdu, dfdx, dgdu, dgdx, dJdu, dJdx = jacs_fun(x, u, d)
        Ju = dJdu - dJdx @ solve(dfdx, dfdu)
        gu = dgdu - dgdx @ solve(dfdx, dfdu)

        if not flag:
            name = 'ss_grad'
            print('Warning! Method {} did not converge. Results are not reliable from here on.'.format(name))
        return {'Ju': Ju, 'gu': gu, 'success': flag}

    def ss_optim(self, d, ignore_flag=False):
        # nx = np.prod(self.m['x'].shape)
        # nz = np.prod(self.m['z'].shape)
        nu = np.prod(self.m['u'].shape)
        nf = np.prod(self.m['f'].shape)
        nh = np.prod(self.m['h'].shape)
        ng = np.prod(self.m['g'].shape)
        nlp = {
            'x': vertcat(self.m['u'], self.m['x'], self.m['z']),
            'p': self.m['p'],
            'f': self.m['J'],
            'g': vertcat(self.m['f'], self.m['h'], self.m['g'])
        }

        u0, lbu, ubu, x0, lbx, ubx, z0, lbz, ubz = self.guess(d)
        x0 = vertcat(u0, x0, z0)
        lbx = vertcat(lbu, lbx, lbz)
        ubx = vertcat(ubu, ubx, ubz)

        lbg = vertcat(DM.zeros(nf+nh), -DM.inf(ng))
        ubg = DM.zeros(nf+nh+ng)

        F = nlpsol('F', 'ipopt', nlp, nlpopts)
        r = F(x0=x0, p=d, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)

        sol = r['x']
        lamg = r['lam_g'][-ng:]
        g = r['g'][-ng:]
        J = r['f']

        if (not F.stats()['success']) and (not ignore_flag):
            name = 'ss_optim'
            print('Warning! Method {} did not converge. Results are not reliable from here on.'.format(name))
        
        dict_sol = {'u': sol[:nu], 'xz': sol[nu:], 'g': g, 'J': J, 'lam_g': lamg, 'success': F.stats()['success']}

        if 'y' in self.m:
            yfun = Function('yf', [self.m['t'], vertcat(self.m['x'], self.m['z']), self.m['u'], self.m['p']], [self.m['y']])
            y = yfun(0, sol[nu:], sol[:nu], d)
            dict_sol['y'] = y

        return dict_sol

    def ss_output(self, u, d, xzguess=None):

        #print(u)
        # nx = np.prod(self.m['x'].shape)
        # nz = np.prod(self.m['z'].shape)
        nu = np.prod(self.m['u'].shape)
        nf = np.prod(self.m['f'].shape)
        nh = np.prod(self.m['h'].shape)
        nlp = {
            'x': vertcat(self.m['u'], self.m['x'], self.m['z']),
            'p': self.m['p'],
            'f': self.m['J'],
            'g': vertcat(self.m['f'], self.m['h'])
        }

        _, _, _, x0, lbx, ubx, z0, lbz, ubz = self.guess(d)
        x0 = vertcat(u, *((x0, z0) if xzguess is None else (xzguess,) ))
        lbx = vertcat(u, lbx, lbz)
        ubx = vertcat(u, ubx, ubz)

        lbg = DM.zeros(nf+nh)
        ubg = DM.zeros(nf+nh)

        F = nlpsol('F', 'ipopt', nlp, nlpopts)
        r = F(x0=x0, p=d, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)

        if not F.stats()['success']:
            # don't give up yet! maybe optimal solution is a good guess?
            optsol = self.ss_optim(d, ignore_flag = True)
            x0 = vertcat(optsol['u'], optsol['xz'])
            r = F(x0=x0, p=d, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)

        sol = r['x']
        J = r['f']

        # Fz = Function('F',[vertcat(self.m['x'], self.m['z']), self.m['u'], self.m['p']],[vertcat(self.m['f'], self.m['h']), self.m['J'], self.m['g'], self.m['y']], ['xz','u','p'], ['f','J','g','y'])
        # rf = rootfinder('G','newton',Fz, {'error_on_fail': False})
        # xsol, Jsol, gsol, ysol = rf(sol[nu:], sol[:nu], d)

        if not F.stats()['success']:
            name = 'ss_output'
            print('Warning! Method {} did not converge. Results are not reliable from here on.'.format(name))

        dict_sol = {'u': sol[:nu], 'xz': sol[nu:], 'J': J, 'success': F.stats()['success']}

        if 'y' in self.m:
            outfun = Function('yf', [self.m['t'], vertcat(self.m['x'], self.m['z']), self.m['u'], self.m['p']], [self.m['y']])
            y = outfun(0, sol[nu:], u, d)
            dict_sol['y'] = y
        if 'g' in self.m:
            outfun = Function('yf', [self.m['t'], vertcat(self.m['x'], self.m['z']), self.m['u'], self.m['p']], [self.m['g']])
            g = outfun(0, sol[nu:], u, d)
            dict_sol['g'] = g

        return dict_sol

class SSModel:
    def __init__(self):
        self.m = {
            'x': DM([]),
            'u': DM([]),
            'p': DM([]),
            'f': DM([]),
            'g': DM([]),
            'J': DM([])
        }
        self.jacs = {
            'dfdu': DM([]),
            'dfdx': DM([]),
            'dJdu': DM([]),
            'dJdx': DM([]),
            'dgdu': DM([]),
            'dgdx': DM([])
        }

    def guess(self, d):
        nx = self.m['x'].shape
        nu = self.m['u'].shape

        x0  = DM.zeros(nx)
        u0  = DM.zeros(nu)
        lbx = -DM.inf(nx)
        lbu = -DM.inf(nu)
        ubx = DM.inf(nx)
        ubu = DM.inf(nu)
        return x0, lbx, ubx, u0, lbu, ubu

    def gen_jacs(self):
        self.jacs = {
            'dfdu': jacobian(self.m['f'], self.m['u']),
            'dfdx': jacobian(self.m['f'], self.m['x']),
            'dJdu': jacobian(self.m['J'], self.m['u']),
            'dJdx': jacobian(self.m['J'], self.m['x']),
            'dgdu': jacobian(self.m['g'], self.m['u']),
            'dgdx': jacobian(self.m['g'], self.m['x'])
        }

    def output_fsolve(self, u, d):
        F = Function('F', [self.m['x'], self.m['u'], self.m['p']], [self.m['f'], self.m['g'], self.m['J']])
        G = rootfinder('G','newton',F)
        x0 = self.guess(d)[0]
        x, g, J = G(x0, u, d)
        return {'x': x, 'g': g, 'J': J, 'success': G.stats()['success']}

    def output(self, u, d):
        nx = np.prod(self.m['x'].shape)
        nu = np.prod(self.m['u'].shape)
        nf = np.prod(self.m['f'].shape)
        nlp = {
            'x': vertcat(self.m['u'], self.m['x']),
            'p': self.m['p'],
            'f': self.m['J'],
            'g': self.m['f']
        }

        x0, lbx, ubx, _, _, _ = self.guess(d)
        x0  = vertcat(u, x0)
        lbx = vertcat(u, lbx)
        ubx = vertcat(u, ubx)

        F = nlpsol('F', 'ipopt', nlp, nlpopts)
        r = F(x0=x0, p=d, lbx=lbx, ubx=ubx, lbg=DM.zeros(nf), ubg=DM.zeros(nf))

        sol = r['x']
        J = r['f']
        outf = Function('outf', [nlp['x'], nlp['p']], [self.m['g']])
        g = outf(sol, d)

        return {'x': sol[-nx:], 'g': g, 'J': J, 'success': F.stats()['success']}

    def optim(self, d):
        nx = np.prod(self.m['x'].shape)
        nu = np.prod(self.m['u'].shape)
        nf = np.prod(self.m['f'].shape)
        ng = np.prod(self.m['g'].shape)
        nlp = {
            'x': vertcat(self.m['u'], self.m['x']),
            'p': self.m['p'],
            'f': self.m['J'],
            'g': vertcat(self.m['f'], self.m['g'])
        }

        x0, lbx, ubx, u0, lbu, ubu = self.guess(d)
        x0 = vertcat(u0, x0)
        lbx = vertcat(lbu, lbx)
        ubx = vertcat(ubu, ubx)

        lbg = vertcat(DM.zeros(nf), -DM.inf(ng))
        ubg = DM.zeros(nf+ng)

        F = nlpsol('F', 'ipopt', nlp, nlpopts)
        r = F(x0=x0, p=d, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)

        sol = r['x']
        lamg = r['lam_g'][-ng:]
        g = r['g'][-ng:]
        J = r['f']

        return {'u': sol[:nu], 'x': sol[-nx:], 'g': g, 'J': J, 'lam_g': lamg, 'success': F.stats()['success']}

    def cost(self, u, d, x=None):
        F = Function('F',[self.m['x'], self.m['u'], self.m['p']],[self.m['J'], self.m['g']], ['x','u','p'], ['J','g'])
        if x is None:
            out = self.output(u,d) #np.zeros(nx)
            J, g = F(out['x'], u, d)
            flag = out['success']
        else:
            J, g = F(x, u, d)
            flag = True
        return {'J': J, 'g': g, 'success': flag}
        
    def grad(self, u, d, x=None):
        F = Function('F',[self.m['x'], self.m['u'], self.m['p']],[self.m['f'], self.m['J'], self.m['g']], ['x','u','p'], ['f','J','g'])
        G = rootfinder('G','newton',F, {'error_on_fail': False})
        if x is None:
            out = self.output(u,d) #np.zeros(nx)
            x = out['x']
        gradfun = G.factory('Ju', ['x','u','p'], ['jac:J:u','jac:g:u'])

        _ = G(x, u, d)
        Ju, gu = gradfun(x, u, d)
        return {'Ju': Ju, 'gu': gu, 'success': G.stats()['success']}

    def grad_2(self, u, d, x=None):
        m = self.m
        jacs = self.jacs
        Jufun = Function('Ju', [m['x'], m['u'], m['p']], [jacs['dJdu'] - jacs['dJdx'] @ solve(jacs['dfdx'], jacs['dfdu'])])
        gufun = Function('gu', [m['x'], m['u'], m['p']], [jacs['dgdu'] - jacs['dgdx'] @ solve(jacs['dfdx'], jacs['dfdu'])])
        flag = True
        if x is None:
            out = self.output(u,d) #np.zeros(nx)
            x = out['x']
            flag = out['success']
        Ju = Jufun(x, u, d)
        gu = gufun(x, u, d)
        return {'Ju': Ju, 'gu': gu.full(), 'success': flag}

# class ClosedLoopSSModel:
#     def __init__(self, model):
#         nu = np.prod(self.m['u'].shape)
#         self.m = {
#             't': model.m['t'],
#             'x': model.m['x'],
#             'z': model.m['z'],
#             'u': model.m['u'],
#             'p': model.m['p'],
#             'f': model.m['f'],
#             'h': model.m['h'],
#             'g': model.m['g'],
#             'J': model.m['J']
#         }